from pydantic import BaseModel

from power_platform_security_assessment.base_classes import ConnectorWithConnections


class ConnectorsAnalysisResult(BaseModel):
    deprecated_connectors: list[ConnectorWithConnections]
    untrusted_connectors: list[ConnectorWithConnections]


class ConnectorsAnalysisReport(BaseModel):
    connectors_info: ConnectorsAnalysisResult
    textual_report: str



from urllib.parse import urlencode

from power_platform_security_assessment.base_classes import ModelDrivenApp
from power_platform_security_assessment.fetchers.base_resource_fetcher import BaseResourceFetcher
from power_platform_security_assessment.token_manager import TokenManager


class ModelDrivenAppsFetcher(BaseResourceFetcher):
    def __init__(self, instance_api_url: str, env_id: str, token_manager: TokenManager):
        self._instance_api_url = instance_api_url
        super().__init__(env_id, token_manager)

    def _get_request_url(self):
        params = {
            "$select": "appmoduleidunique,statecode",
        }
        return f'{self._instance_api_url}/api/data/v9.2/appmodules/Microsoft.Dynamics.CRM.RetrieveUnpublishedMultiple()?{urlencode(params)}'

    def _fetch_single_page_model_driven_apps(self, token: str, url: str):
        response_data = self._fetch_single_page(url, headers={
            'Authorization': f'Bearer {token}',
        })

        model_driven_apps = [ModelDrivenApp(**app) for app in response_data.get('value', [])]
        next_page = response_data.get('@odata.nextLink', None)

        return model_driven_apps, next_page

    def _fetch_model_driven_apps(self) -> list[ModelDrivenApp]:
        token = self._token_manager.fetch_access_token(scope=[f'{self._instance_api_url}/.default'])
        next_page_url = self._get_request_url()
        all_model_driven_apps = []

        while next_page_url:
            model_driven_apps, next_page_url = self._fetch_single_page_model_driven_apps(token=token, url=next_page_url)
            all_model_driven_apps.extend(model_driven_apps)
            self._resource_count += len(model_driven_apps)

        return all_model_driven_apps

    def _do_fetch_resource_data(self) -> list[ModelDrivenApp]:
        return self._fetch_model_driven_apps()



from datetime import datetime
from urllib.parse import urlencode

import requests
from dateutil import parser
from dateutil.relativedelta import relativedelta
from pydash import find_index, chain

from power_platform_security_assessment.base_classes import Environment


class EnvironmentsFetcher:
    _MAX_ENVIRONMENTS_TO_SCAN = 10

    def __init__(self):
        self.environments = []

    @staticmethod
    def _display_environments(environments: list[Environment], total_envs: int):
        max_display_name_length = max([len(env.properties.displayName) for env in environments])
        print(f'Total number of environments: {total_envs}')
        print()
        print(
            f'{"ID":<44} {"Name":<{max_display_name_length}} {"Created By":<20} {"Create Time":<30} {"Last Activity":<30} {"Type":<10}')
        for env in environments:
            created_by = env.properties.createdBy.get('displayName', 'N/A')
            print(
                f'{env.id.split("/")[-1]:<44} {env.properties.displayName:<{max_display_name_length}} {created_by:<20} {env.properties.createdTime:<30} {env.properties.lastActivity.lastActivity.lastActivityTime:<30} {env.properties.environmentSku:<10}')
        print()

    def _notify_user(self, total_envs: int):
        if total_envs > self._MAX_ENVIRONMENTS_TO_SCAN:
            print(f'The number of environments exceeds {self._MAX_ENVIRONMENTS_TO_SCAN}. '
                  f'Scanning only the selected environments due to runtime limitations.')

    @staticmethod
    def _get_request_url() -> str:
        url = 'https://api.bap.microsoft.com/providers/Microsoft.BusinessAppPlatform/scopes/admin/environments'
        params = {
            'api-version': '2021-04-01',
            '$expand': 'properties/scheduledLifecycleOperations',
            '$select': 'id,type,name,properties.displayName,properties.createdTime,properties.environmentSku,'
                       'properties.createdBy.displayName,properties.lastActivity.lastActivity.lastActivityTime,'
                       'properties.isDefault,properties.linkedEnvironmentMetadata.instanceApiUrl'
        }
        return f'{url}?{urlencode(params)}'

    @staticmethod
    def _fetch_single_page_environments(token: str, url: str) -> tuple[list[Environment], str]:
        res = requests.get(
            url,
            headers={'Authorization': f'Bearer {token}'}
        )
        response_data = res.json()

        environments = [Environment(**env) for env in response_data.get('value', [])]
        next_page = response_data.get('nextLink', None)

        return environments, next_page

    def _fetch_environments(self, token: str) -> list[Environment]:
        next_page_url = self._get_request_url()
        all_envs = []

        while next_page_url:
            envs, next_page_url = self._fetch_single_page_environments(token=token, url=next_page_url)
            all_envs.extend(envs)

        return all_envs

    @staticmethod
    def _compare_environments(env1: Environment, env2: Environment):
        # Default environment > Production > Developer > Sandbox > Other
        env_types = ['Sandbox', 'Developer', 'Production', 'Default']
        env1_type = find_index(env_types, lambda env_type: env_type == env1.properties.environmentSku)
        env2_type = find_index(env_types, lambda env_type: env_type == env2.properties.environmentSku)
        if env2_type != env1_type:
            return env2_type - env1_type

        # If both are of the same type, prefer the one that has an activity in the last month
        threshold_date = (datetime.now() - relativedelta(days=30)).timestamp()
        env1_last_activity = parser.isoparse(env1.properties.lastActivity.lastActivity.lastActivityTime).timestamp()
        env2_last_activity = parser.isoparse(env2.properties.lastActivity.lastActivity.lastActivityTime).timestamp()
        env1_last_activity_in_last_month = env1_last_activity > threshold_date
        env2_last_activity_in_last_month = env2_last_activity > threshold_date
        if env2_last_activity_in_last_month != env1_last_activity_in_last_month:
            return env2_last_activity_in_last_month - env1_last_activity_in_last_month

        # If both have activity in the last month, prefer the one that was created earlier
        env1_created_time = parser.isoparse(env1.properties.createdTime).timestamp()
        env2_created_time = parser.isoparse(env2.properties.createdTime).timestamp()
        return env1_created_time - env2_created_time

    def fetch_environments(self, token) -> tuple[list[Environment], int]:
        environments = self._fetch_environments(token)
        total_envs = len(environments)
        self._notify_user(total_envs)
        selected_environments: list[Environment] = list(
            chain(environments)
            .sort(self._compare_environments)
            .take(self._MAX_ENVIRONMENTS_TO_SCAN)
            .value()
        )
        self._display_environments(selected_environments, total_envs)
        return selected_environments, total_envs



from power_platform_security_assessment.base_classes import Application


class BypassConsentTextualReport:

    @staticmethod
    def generate_textual_report(bypass_consent_apps: list[Application]) -> str:
        verb = "is" if len(bypass_consent_apps) == 1 else "are"
        has_or_have = "has" if len(bypass_consent_apps) == 1 else "have"

        textual_report = (
            f'There {verb} <b>{len(bypass_consent_apps)}</b> application{"" if len(bypass_consent_apps) == 1 else "s"} '
            f'that {has_or_have} the <b>"Bypass Consent"</b> feature enabled.<br>'
            f'This feature allows the application owner to access or change information on behalf of '
            f'the application users, without their consent.'
        )

        return textual_report



from typing import Dict, Any, Optional, TypeVar, Generic

from pydantic import BaseModel

T = TypeVar("T")


class LastActivityTimes(BaseModel):
    lastActivityTime: str


class LastActivity(BaseModel):
    lastActivity: LastActivityTimes


class LinkedEnvironmentMetadata(BaseModel):
    instanceApiUrl: str


class EnvironmentProperties(BaseModel):
    displayName: str
    createdTime: str
    createdBy: Dict[str, Any]
    lastActivity: LastActivity
    environmentSku: str
    isDefault: bool
    linkedEnvironmentMetadata: Optional[LinkedEnvironmentMetadata] = None


class Environment(BaseModel):
    id: str
    type: str
    name: str
    properties: EnvironmentProperties


class ApplicationUser(BaseModel):
    id: str


class DataLossPreventionEvaluationResult(BaseModel):
    violations: Optional[list] = []


class AppQuarantineState(BaseModel):
    quarantineStatus: str


class AppExecutionRestrictions(BaseModel):
    dataLossPreventionEvaluationResult: Optional[DataLossPreventionEvaluationResult] = None
    appQuarantineState: Optional[AppQuarantineState] = None


class EmbeddedApp(BaseModel):
    type: str


class ApplicationProperties(BaseModel):
    appVersion: str
    createdTime: str
    lastModifiedTime: str
    sharedGroupsCount: int
    sharedUsersCount: int
    displayName: str
    bypassConsent: bool
    owner: ApplicationUser
    createdBy: ApplicationUser
    executionRestrictions: Optional[AppExecutionRestrictions] = None
    embeddedApp: Optional[EmbeddedApp] = None


class Application(BaseModel):
    id: str
    name: str
    logicalName: Optional[str] = None
    type: str
    appType: str
    properties: ApplicationProperties


class CloudFlowUser(BaseModel):
    userId: Optional[str] = None


class CloudFlowProperties(BaseModel):
    displayName: str
    createdTime: str
    lastModifiedTime: str
    state: str
    workflowEntityId: Optional[str] = None
    creator: CloudFlowUser


class CloudFlow(BaseModel):
    id: str
    name: str
    type: str
    properties: CloudFlowProperties


class DesktopFlow(BaseModel):
    workflowidunique: str
    statecode: int


class ModelDrivenApp(BaseModel):
    appmoduleidunique: str
    statecode: int


class User(BaseModel):
    domainname: str
    isdisabled: bool
    azurestate: int
    fullname: str
    azureactivedirectoryobjectid: str


class ConnectorMetadata(BaseModel):
    source: Optional[str] = None


class ConnectorProperties(BaseModel):
    displayName: str
    publisher: str
    metadata: ConnectorMetadata


class Connector(BaseModel):
    name: str
    properties: ConnectorProperties


class Connection(BaseModel):
    name: str
    id: str


class ConnectorWithConnections(BaseModel):
    connector: Connector
    connections: list[Connection]


class ResourceData(BaseModel, Generic[T]):
    value: list[T] = []
    count: int = 0
    all_resources_fetched: bool = True



from typing import List

import msal

from power_platform_security_assessment.consts import ResponseKeys


class TokenManager:
    def __init__(self, client_id: str, refresh_token: str):
        self._client_id = client_id
        self._refresh_token = refresh_token

    def fetch_access_token(self, scope: List[str]) -> str:
        app = msal.PublicClientApplication(self._client_id)
        result = app.acquire_token_by_refresh_token(self._refresh_token, scopes=scope)
        if ResponseKeys.ACCESS_TOKEN in result:
            return result[ResponseKeys.ACCESS_TOKEN]
        else:
            raise Exception("Failed to acquire token: %s" % result.get("error_description"))



from pydantic import BaseModel

from power_platform_security_assessment.base_classes import Application


class BypassConsentResult(BaseModel):
    bypass_consent_apps: list[Application]
    textual_report: str



from setuptools import setup, find_packages

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

with open('README.md') as f:
    long_description = f.read()

setup(
    name='power-platform-security-assessment',
    version='0.1.0',
    description='Power Platform Security Assessment Tool',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Nokod Security',
    author_email='support@nokodsecurity.com',
    url='https://github.com/Nokod/power-platform-security-assessment',
    packages=find_packages(),
    install_requires=requirements,
    include_package_data=True,
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'power-platform-security-assessment=power_platform_security_assessment.security_assessment_tool:main',
        ],
    },
)



from pydash import map_values, key_by, sort_by, values

from power_platform_security_assessment.base_classes import CloudFlow, Application, User, Environment
from power_platform_security_assessment.security_features.app_developers.app_developer_textual_report import AppDeveloperTextualReport
from power_platform_security_assessment.security_features.app_developers.model import UserResources, AppDevelopersReport, Developers
from power_platform_security_assessment.security_features.common import get_application_owner_id, get_cloud_flow_owner_id


class AppDeveloperAnalyzer:
    def __init__(self, apps: list[Application], cloud_flows: list[CloudFlow], users: list[User], environments: list[Environment]):
        self._apps = apps
        self._cloud_flows = cloud_flows
        self._users = users
        self._textual_report_generator = AppDeveloperTextualReport(environments)

    def _get_user_to_apps_and_flows_map(self, users: list[User]) -> list[UserResources]:
        # Create a map of user_id → UserResources
        user_map: dict[str, UserResources] = map_values(
            key_by(users, lambda user: user.azureactivedirectoryobjectid),
            lambda user: UserResources(
                user=user,
                apps=[],
                flows=[],
            )
        )

        # Map apps to their owners
        for app in self._apps:
            owner_id = get_application_owner_id(app)
            if owner_id in user_map:
                user_map[owner_id].apps.append(app)

        # Map flows to their creators
        for flow in self._cloud_flows:
            owner_id = get_cloud_flow_owner_id(flow)
            if owner_id in user_map:
                user_map[owner_id].flows.append(flow)

        # Convert the map to a list and sort it by the number of apps and flows
        return sort_by(
            values(user_map),
            lambda user_resources: len(user_resources.apps) + len(user_resources.flows),
            reverse=True,
        )

    def analyze(self) -> AppDevelopersReport:
        guest_users = []
        deleted_users = []

        for user in self._users:
            if '#EXT#' in user.domainname:
                guest_users.append(user)
            if user.azurestate in {1, 2}:
                deleted_users.append(user)

        guest_developers = self._get_user_to_apps_and_flows_map(guest_users)
        inactive_developers = self._get_user_to_apps_and_flows_map(deleted_users)

        developers_info = Developers(
            guest_developers=guest_developers,
            inactive_developers=inactive_developers,
        )
        textural_report = self._textual_report_generator.generate_textual_report(developers_info)

        return AppDevelopersReport(
            developers_info=developers_info,
            textual_report=textural_report,
        )



from urllib.parse import urlencode

from power_platform_security_assessment.base_classes import DesktopFlow
from power_platform_security_assessment.fetchers.base_resource_fetcher import BaseResourceFetcher
from power_platform_security_assessment.token_manager import TokenManager


class DesktopFlowsFetcher(BaseResourceFetcher):
    def __init__(self, instance_api_url: str, env_id: str, token_manager: TokenManager):
        self._instance_api_url = instance_api_url
        super().__init__(env_id, token_manager)

    def _get_request_url(self):
        params = {
            "$select": "name,workflowidunique,statecode",
            "$count": "true",
            "$filter": "category eq 6",
        }
        return f'{self._instance_api_url}/api/data/v9.2/workflows?{urlencode(params)}'

    def _fetch_single_page_desktop_flows(self, token: str, url: str):
        response_data = self._fetch_single_page(url, headers={
            'Authorization': f'Bearer {token}',
        })

        desktop_flows = [DesktopFlow(**flow) for flow in response_data.get('value', [])]
        next_page = response_data.get('@odata.nextLink', None)

        return desktop_flows, next_page

    def _fetch_desktop_flows(self) -> list[DesktopFlow]:
        token = self._token_manager.fetch_access_token(scope=[f'{self._instance_api_url}/.default'])
        next_page_url = self._get_request_url()
        all_desktop_flows = []

        while next_page_url:
            desktop_flows, next_page_url = self._fetch_single_page_desktop_flows(token=token, url=next_page_url)
            all_desktop_flows.extend(desktop_flows)
            self._resource_count += len(desktop_flows)

        return all_desktop_flows

    def _do_fetch_resource_data(self) -> list[DesktopFlow]:
        return self._fetch_desktop_flows()



from pydash import get

from power_platform_security_assessment.base_classes import ConnectorWithConnections
from power_platform_security_assessment.security_features.connectors.connectors_analyzer_textual_report import (
    ConnectorsAnalyzerTextualReport
)
from power_platform_security_assessment.security_features.connectors.deprecated_connectors import DEPRECATED_CONNECTOR_NAMES
from power_platform_security_assessment.security_features.connectors.model import ConnectorsAnalysisResult, ConnectorsAnalysisReport


class ConnectorsAnalyzer:
    def __init__(self, connectors_with_connections: list[ConnectorWithConnections]):
        self._connectors_with_connections = connectors_with_connections
        self._connectors_analysis_textual_report_generator = ConnectorsAnalyzerTextualReport()

    def analyze(self) -> ConnectorsAnalysisReport:
        deprecated_connectors = [
            c for c in self._connectors_with_connections
            if c.connector.properties.displayName.endswith(' (Deprecated)')
               or c.connector.properties.displayName in DEPRECATED_CONNECTOR_NAMES
        ]

        untrusted_connectors = [
            c for c in self._connectors_with_connections
            if get(c.connector.properties.metadata, 'source') == 'independentpublisher'
        ]

        connectors_analysis_result = ConnectorsAnalysisResult(
            deprecated_connectors=deprecated_connectors,
            untrusted_connectors=untrusted_connectors,
        )
        textual_report = self._connectors_analysis_textual_report_generator.generate_textual_report(connectors_analysis_result)

        return ConnectorsAnalysisReport(
            connectors_info=connectors_analysis_result,
            textual_report=textual_report,
        )



import os

import pandas as pd
import plotly.graph_objects as go
from jinja2 import Template

from power_platform_security_assessment.base_classes import User, CloudFlow, Application, ModelDrivenApp, DesktopFlow, \
    ConnectorWithConnections, Environment, ResourceData
from power_platform_security_assessment.consts import ComponentType
from power_platform_security_assessment.security_features.common import is_app_disabled, is_flow_disabled, \
    is_model_driven_app_disabled, is_desktop_flow_disabled
from power_platform_security_assessment.utils import get_environment_developers_count, round_time_to_seconds


class ReportBuilder:
    BACKGROUND_COLOR = '#f9fafb'
    TITLE_COLOR = '#b0b7c5'
    MAIN_BLUE = '#688bf8'
    MAIN_GRAY = '#62708b'
    MAIN_GREEN = '#50caa9'
    MAIN_RED = '#ff4b4b'
    MAIN_ORANGE = '#f9a14d'
    VULN_COLOR = '#9c9c9c'

    def __init__(self, applications: list[Application], cloud_flows: list[CloudFlow], desktop_flows: list[DesktopFlow],
                 model_driven_apps: list[ModelDrivenApp], users: list[User],
                 connectors: list[ConnectorWithConnections], environments_results: list, failed_environments: list,
                 environments: list[Environment], total_envs: int, all_users_fetched: bool):
        self._applications = applications
        self._cloud_flows = cloud_flows
        self._desktop_flows = desktop_flows
        self._model_driven_apps = model_driven_apps
        self._users = users
        self._connectors = connectors
        self._scanned_environments = environments
        self._environments_results = environments_results
        self._failed_environments = failed_environments
        self._total_envs_count = total_envs
        self._all_users_fetched = all_users_fetched

    def build_report(self, app_developers_report, connectors_report, bypass_consent_report):
        env_summary = [self._build_env_summary()]
        environment_reports = [
            self._build_bar_chart_summary(),
            self._build_components_in_env(),
            self._build_connectors_in_env(),
        ]
        user_reports = [self._build_users_pie_charts()]
        top_3_reports = [self._build_biggest_environments(), self._build_used_connections()]
        email_body = self._build_email_body()
        __location__ = os.path.realpath(
            os.path.join(os.getcwd(), os.path.dirname(__file__)))

        with open(os.path.join(__location__, 'report.html')) as f:
            template = Template(f.read())

        rendered_template = template.render(
            found_environments=env_summary,
            environment_reports=environment_reports,
            total_users=f'{len(self._users)}' + ("+" if not self._all_users_fetched else ""),
            user_reports=user_reports,
            app_developers_report=app_developers_report,
            connectors_report=connectors_report,
            bypass_consent_report=bypass_consent_report,
            top_3_reports=top_3_reports,
            email_body=email_body,
            report_date=pd.Timestamp.now().strftime('%B %d, %Y'),
            total_envs_count=self._total_envs_count,
        )
        with open('power_platform_scan_report.html', 'w') as f:
            f.write(rendered_template)
            print(f'Report generated successfully. Output saved to {os.path.abspath("power_platform_scan_report.html")}')

    def _build_email_body(self):
        data = self.get_components_per_env()[:-1]
        output = [
            f'Hello Nokod Team,\\n'
            f'I would like to discuss the results of the scan I made.\\n'
            f'See below findings.\\n'
        ]
        for i, env in enumerate(data):
            output.append(
                f'Environment \'Env{i + 1}\' has {env["Canvas Apps"]} applications, {env["Cloud Flows"]} cloud '
                f'flows, {env["Desktop Flows"]} desktop flows, and {env["Model Driven Apps"]} model driven apps. '
                f'This totals to {env["Total"]} components.')
        return "\\n".join(output)

    def _build_env_summary(self):
        envs = self.get_scanned_environments()
        df = pd.DataFrame(envs)
        fig = go.Figure(data=[go.Table(
            header=dict(values=list(df.columns),
                        fill_color=self.TITLE_COLOR,
                        align='left'),
            cells=dict(values=[df['Name'], df['Type'], df['Created By'], df['Create Time'], df['Last Activity'],
                               df['Scan Status']],
                       fill_color=self.BACKGROUND_COLOR,
                       line=dict(color='white'),
                       align='left'))],
            layout={
                'plot_bgcolor': self.BACKGROUND_COLOR,
                'width': 1400,
                'height': 150 + (len(envs) * 20),
                'margin': dict(l=0, r=0, t=50, b=50),
                'annotations': [
                    go.layout.Annotation(
                        x=0, y=-0.1, showarrow=False, xanchor='left', yanchor='bottom',
                        text=f'Total number of environments: {self._total_envs_count}. '
                    ),
                ],
            }
        )

        config = {'displaylogo': False, 'modeBarButtonsToRemove': ['toImage']}
        return fig.to_html(full_html=False, include_plotlyjs='cdn', config=config)

    def get_scanned_environments(self):
        envs = []
        for env in self._scanned_environments:
            envs.append(self._create_env_data(
                env=env,
                failed=env.name in [env[ComponentType.ENVIRONMENT].name for env in self._failed_environments]
            ))

        return envs

    @staticmethod
    def _create_env_data(env: Environment, failed=False):
        env_name = env.properties.displayName
        created_by = env.properties.createdBy.get('displayName', 'N/A')
        created_time = round_time_to_seconds(env.properties.createdTime)
        last_activity = round_time_to_seconds(env.properties.lastActivity.lastActivity.lastActivityTime)
        env_type = env.properties.environmentSku
        status = 'Failed - insufficient permissions' if failed else 'Success'
        env_data = {'Name': env_name, 'Type': env_type, 'Created By': created_by, 'Create Time': created_time,
                    'Last Activity': last_activity, 'Scan Status': status}
        return env_data

    def _build_components_in_env(self):
        data = self.get_components_per_env()
        df = pd.DataFrame(data)
        fig = go.Figure(data=[go.Table(
            header=dict(
                values=[
                    'Name',
                    '<a href="https://learn.microsoft.com/en-us/power-apps/maker/canvas-apps/getting-started">Canvas Apps</a>',
                    '<a href="https://learn.microsoft.com/en-us/power-automate/overview-cloud">Cloud Flows</a>',
                    '<a href="https://learn.microsoft.com/en-us/power-automate/desktop-flows/introduction">Desktop Flows</a>',
                    '<a href="https://learn.microsoft.com/en-us/power-apps/maker/model-driven-apps/model-driven-app-overview">Model Driven Apps</a>',
                    'Total'
                ],
                fill_color=self.TITLE_COLOR,
                align='left'
            ),
            cells=dict(values=[df['Name'], df['Canvas Apps'], df['Cloud Flows'], df['Desktop Flows'],
                               df['Model Driven Apps'], df['Total']],
                       fill_color=self.BACKGROUND_COLOR,
                       align='left'))],
            layout={'title': {'text': 'Components per Environment', 'x': 0.5}, 'plot_bgcolor': self.BACKGROUND_COLOR,
                    'margin': dict(l=10, r=10, t=100, b=0)})
        config = {'displaylogo': False, 'modeBarButtonsToRemove': ['toImage']}
        return fig.to_html(full_html=False, include_plotlyjs='cdn', config=config)

    @staticmethod
    def _get_component_count(resource_data: ResourceData) -> tuple:
        count = resource_data.count
        display = f'{count}' + ("+" if not resource_data.all_resources_fetched else "")
        return count, display

    def get_components_per_env(self):
        data = []
        for env in self._environments_results:
            canvas_count, canvas_display = self._get_component_count(env[ComponentType.APPLICATIONS])
            cloud_count, cloud_display = self._get_component_count(env[ComponentType.CLOUD_FLOWS])
            desktop_count, desktop_display = self._get_component_count(env[ComponentType.DESKTOP_FLOWS])
            model_count, model_display = self._get_component_count(env[ComponentType.MODEL_DRIVEN_APPS])
            total_count = canvas_count + cloud_count + desktop_count + model_count
            total_display = f'{total_count}' + ("+" if not all([env[ComponentType.APPLICATIONS].all_resources_fetched,
                                                                env[ComponentType.CLOUD_FLOWS].all_resources_fetched,
                                                                env[ComponentType.DESKTOP_FLOWS].all_resources_fetched,
                                                                env[ComponentType.MODEL_DRIVEN_APPS].all_resources_fetched]) else "")
            data.append({
                'Name': env[ComponentType.ENVIRONMENT].properties.displayName,
                'Canvas Apps': canvas_display,
                'Cloud Flows': cloud_display,
                'Desktop Flows': desktop_display,
                'Model Driven Apps': model_display,
                'Total': total_display
            })

        rows = sorted(data, key=lambda x: int(x['Total'].rstrip('+')), reverse=True)
        total_row = {
            'Name': '<b>Total</b>',
            'Canvas Apps': f'<b>{sum(int(row["Canvas Apps"].rstrip("+")) for row in rows)}</b>',
            'Cloud Flows': f'<b>{sum(int(row["Cloud Flows"].rstrip("+")) for row in rows)}</b>',
            'Desktop Flows': f'<b>{sum(int(row["Desktop Flows"].rstrip("+")) for row in rows)}</b>',
            'Model Driven Apps': f'<b>{sum(int(row["Model Driven Apps"].rstrip("+")) for row in rows)}</b>',
            'Total': f'<b>{sum(int(row["Total"].rstrip("+")) for row in rows)}</b>'
        }

        return rows + [total_row]

    def _build_biggest_environments(self):
        envs = self._get_biggest_environments()
        df = pd.DataFrame(envs[:3])
        fig = go.Figure(data=[go.Table(
            header=dict(values=list(df.columns),
                        fill_color=self.TITLE_COLOR,
                        align='left'),
            cells=dict(values=[df['Environment Name'], df['Type'], df['Total Components'], df['Developers *']],
                       fill_color=self.BACKGROUND_COLOR,
                       align='left'))],
            layout={'title': {'text': 'Top 3 Biggest Environments', 'x': 0.5}, 'height': 300, 'annotations': [
                go.layout.Annotation(x=0, y=-0.2, showarrow=False, xanchor='left', yanchor='bottom',
                                     text='* Developers are users who own at least one application in the '
                                          'environment')]})
        config = {'displaylogo': False, 'modeBarButtonsToRemove': ['toImage']}
        return fig.to_html(full_html=False, include_plotlyjs='cdn', config=config)

    def _get_biggest_environments(self):
        envs = []
        for env in self._environments_results:
            env_name = env[ComponentType.ENVIRONMENT].properties.displayName
            apps_count = env[ComponentType.APPLICATIONS].count
            flows_count = env[ComponentType.CLOUD_FLOWS].count
            model_driven_apps_count = env[ComponentType.MODEL_DRIVEN_APPS].count
            desktop_flows_count = env[ComponentType.DESKTOP_FLOWS].count
            total_components = apps_count + flows_count + model_driven_apps_count + desktop_flows_count
            envs.append({'Environment Name': env_name, 'Type': env[ComponentType.ENVIRONMENT].properties.environmentSku,
                         'Total Components': total_components, 'Developers *': get_environment_developers_count(env)})
        envs = sorted(envs, key=lambda x: x['Total Components'], reverse=True)
        return envs

    def _build_used_connections(self):
        connectors = self._get_used_connections()

        df = pd.DataFrame(connectors)
        fig = go.Figure(data=[go.Table(
            header=dict(values=list(df.columns),
                        fill_color=self.TITLE_COLOR,
                        align='left'),
            cells=dict(values=[df['Name'], df['Publisher'], df['Connections Count']],
                       fill_color=self.BACKGROUND_COLOR,
                       align='left'))],
            layout={'title': {'text': 'Top 3 Most Used Connectors', 'x': 0.5}, 'height': 300})

        config = {'displaylogo': False, 'modeBarButtonsToRemove': ['toImage']}
        return fig.to_html(full_html=False, include_plotlyjs='cdn', config=config)

    def _get_used_connections(self):
        connectors = []
        for connector_with_connections in self._connectors[:3]:
            connector = connector_with_connections.connector
            connections_count = len(connector_with_connections.connections)
            connectors.append({'Name': connector.properties.displayName, 'Publisher': connector.properties.publisher,
                               'Connections Count': connections_count})
        sorted(connectors, key=lambda x: x['Connections Count'])
        return connectors

    def _build_connectors_in_env(self):
        data = self._get_connectors_in_env_data()
        df = pd.DataFrame(data)
        fig = go.Figure(
            data=[go.Table(
                header=dict(values=list(df.columns),
                            fill_color=self.TITLE_COLOR,
                            align='left'),
                cells=dict(values=[df['Environment Name'], df['Number of Connectors']],
                           fill_color=self.BACKGROUND_COLOR,
                           align='left'))],
            layout={'title': {'text': 'Connectors per Environment', 'x': 0.5},
                    'plot_bgcolor': self.BACKGROUND_COLOR, 'height': 150 + (len(data) * 20), 'margin': dict(l=10, r=10, t=100, b=0)})
        config = {'displaylogo': False, 'modeBarButtonsToRemove': ['toImage']}
        return fig.to_html(full_html=False, include_plotlyjs='cdn', config=config)

    def _get_connectors_in_env_data(self):
        data = [{'Environment Name': env[ComponentType.ENVIRONMENT].properties.displayName, 'Number of Connectors': len(
            env[ComponentType.CONNECTIONS].value)} for env in self._environments_results]
        data = sorted(data, key=lambda x: x['Number of Connectors'], reverse=True)
        return data

    def _build_users_pie_charts(self):
        df1, df2, df3 = self._get_pie_charts()

        fig1 = go.Figure(
            data=[go.Pie(labels=df1['Type'], values=df1['Count'], name='Internal vs Guest Users',
                         marker=dict(colors=[self.MAIN_GREEN, self.MAIN_ORANGE]))],
            layout={'title': {'text': 'Internal vs Guest Users', 'x': 0.5}, 'height': 300, 'width': 350}
        )
        fig2 = go.Figure(
            data=[go.Pie(labels=df2['Type'], values=df2['Count'], name='Enabled vs Disabled Users',
                         marker=dict(colors=[self.MAIN_GREEN, self.MAIN_RED]))],
            layout={'title': {'text': 'Enabled vs Disabled Users', 'x': 0.5}, 'height': 300, 'width': 350}
        )
        fig3 = go.Figure(
            data=[go.Pie(labels=df3['Type'], values=df3['Count'], name='Users in Active Directory',
                         marker=dict(colors=[self.MAIN_GREEN, self.MAIN_ORANGE, self.MAIN_RED]))],
            layout={'title': {'text': 'Users in Active Directory', 'x': 0.5}, 'height': 300, 'width': 350}
        )

        config = {'displaylogo': False, 'modeBarButtonsToRemove': ['toImage']}

        return (fig1.to_html(full_html=False, include_plotlyjs='cdn', config=config)
                + fig2.to_html(full_html=False, include_plotlyjs='cdn', config=config)
                + fig3.to_html(full_html=False, include_plotlyjs='cdn', config=config))

    def _get_pie_charts(self):
        guest_users = len([user for user in self._users if user.domainname.find('#EXT#') != -1])
        enabled_users = len([user for user in self._users if not user.isdisabled])
        disabled_users = len([user for user in self._users if user.isdisabled])
        azure_state_0 = len([user for user in self._users if not user.azurestate])
        azure_state_1 = len([user for user in self._users if user.azurestate == 1])
        azure_state_2 = len([user for user in self._users if user.azurestate == 2])
        df1 = pd.DataFrame(
            {'Type': ['Internal Users', 'Guest Users'], 'Count': [len(self._users) - guest_users, guest_users]})
        df2 = pd.DataFrame({'Type': ['Enabled Users', 'Disabled Users'], 'Count': [enabled_users, disabled_users]})
        df3 = pd.DataFrame({'Type': ['Active', 'AD Soft Delete', 'AD Hard Delete'],
                            'Count': [azure_state_0, azure_state_1, azure_state_2]})
        return df1, df2, df3

    def _build_bar_chart_summary(self):
        main_green = '#50caa9'
        main_red = '#ff4b4b'

        df = self._get_bar_chart_data()
        fig = go.Figure(
            data=[
                go.Bar(name='Enabled', x=df.columns, y=df.iloc[0], marker=dict(color=main_green)),
                go.Bar(name='Disabled', x=df.columns, y=df.iloc[1], marker=dict(color=main_red))
            ],
            layout={
                'barmode': 'stack',
                'title': {'text': 'Total Components per Type', 'x': 0.5},
                'plot_bgcolor': self.BACKGROUND_COLOR,
            }
        )
        config = {'displaylogo': False, 'modeBarButtonsToRemove': ['toImage']}
        return fig.to_html(full_html=False, include_plotlyjs='cdn', config=config)

    def _get_bar_chart_data(self):
        enabled_apps = len([app for app in self._applications if not is_app_disabled(app)])
        enabled_cloud_flows = len([flow for flow in self._cloud_flows if not is_flow_disabled(flow)])
        enabled_desktop_flows = len([flow for flow in self._desktop_flows if not is_desktop_flow_disabled(flow)])
        enabled_model_driven_apps = len(
            [app for app in self._model_driven_apps if not is_model_driven_app_disabled(app)])
        df = pd.DataFrame({
            'Canvas Apps': [enabled_apps, len(self._applications) - enabled_apps],
            'Cloud Flows': [enabled_cloud_flows, len(self._cloud_flows) - enabled_cloud_flows],
            'Desktop Flows': [enabled_desktop_flows, len(self._desktop_flows) - enabled_desktop_flows],
            'Model Driven Apps': [enabled_model_driven_apps, len(self._model_driven_apps) - enabled_model_driven_apps],
        })
        return df
